/*
 *	taka's thinking routine
 *		for the Weakest Othello League
 *
 *	$Header: /home/taka/prog/oth/RCS/taka.cc,v 1.27 1997-10-24 00:49:23+09 taka Exp $
 */

#define TAKADEBUG 0
#define Thinker taka2Think1


#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#if TAKADEBUG
#include <iostream.h>
#include <iomanip.h>
#endif

#include "othello.h"


/*----------------------------------------------------------------------*/
#if TAKADEBUG*0
struct _new_watcher {
	static long int count;
	_new_watcher() {;}
	~_new_watcher();
} _Anew_watcher;
long int _new_watcher::count=0;
_new_watcher::~_new_watcher()
{
	if (count) {
		cout <<"new/delete count = " <<count <<endl;
	}
}
void *operator new(size_t siz)
{
	_new_watcher::count++;
	return malloc(siz);
}
void operator delete(void *p)
{
	_new_watcher::count--;
	if (p!=NULL) free(p);
}
#endif	/* for TAKADEBUG */


/*----------------------------------------------------------------------*/
/*
 * constants and macros and type definitions
 */
#define TABLE_SIZE 91
#define tbl(x, y) ((y)*9 + (x))
#define tblx(pos) ((pos)%9)
#define tbly(pos) ((pos)/9)
#define tblresult(pos) (tbly(pos)*8 + tblx(pos) -9)

enum { P_WALL = 2, P_SPACE = 0, P_MY_COLOR = 1, P_OP_COLOR = -1 };
typedef signed char cint_t;
typedef cint_t piece_t;
typedef cint_t getlist_t;

#define ESTABLE_FILENAME "taka2.dat"

/* for debug use */
#if TAKADEBUG
#define STATIC
#define DEADBEEF 0x34810567
#else
#define STATIC static
#endif


/*----------------------------------------------------------------------*/

struct taka2Malloc_block_header;

struct taka2Malloc_block_control {
	void *free;
	unsigned int block_size;
	unsigned int allocated_blocks;
	struct taka2Malloc_block_header *block_root;
};

struct taka2Malloc_block_header {
	struct taka2Malloc_block_header *next;
	unsigned int block_size;
	unsigned int allocated_blocks;
	struct taka2Malloc_block_control *pseudo_this;
};


STATIC void *Malloc_block_morecore(struct taka2Malloc_block_control *mbc)
{
	struct taka2Malloc_block_header *mbh;
	unsigned int l,s;
	void **p;

	s=mbc->block_size;
	l=mbc->allocated_blocks;
	mbh=(struct taka2Malloc_block_header *)
	  malloc(sizeof(struct taka2Malloc_block_header)+s*l);
	if (!mbh) return NULL;
	mbh->next=mbc->block_root;
	mbh->block_size=s;
	mbh->allocated_blocks=l;
	mbh->pseudo_this=mbc;
	mbc->block_root=mbh;
	p=(void **)(mbh+1);
#if 0
	while (--l) { *p=(unsigned int)p+s; p=(unsigned int *)((unsigned int)p+s); }
	*p=NULL;
#else
	{
		unsigned int d, nd;
		d = 0;
		while (--l) {
			nd = (d+23)%mbc->allocated_blocks;
			*(void **)((unsigned int)p + d*mbc->block_size) =
			  (void *)((unsigned int)p + nd*mbc->block_size);
			d = nd;
		}
		*(void **)((unsigned int)p + d*mbc->block_size) = NULL;
	}
#endif
	p=(void **)(mbh+1);
	mbc->free=(void *)*p;
	return p;
}

inline STATIC void *
Malloc_block(struct taka2Malloc_block_control *mbc)
{
	unsigned int *p;

	p=(unsigned int *)mbc->free;
	if (p) {
		mbc->free=(void *)*p;
		return p;
	}else {
		return Malloc_block_morecore(mbc);
	}
}

inline STATIC void
Free_block(struct taka2Malloc_block_control *mbc,void *block)
{
	*(unsigned int *)block=(unsigned int)mbc->free;
	mbc->free=block;
}

STATIC struct taka2Malloc_block_control Malloc256 =
        { 0, 256/*512*/+8, 1024*2, 0 };


/*----------------------------------------------------------------------*/
/*
 * class definitions
 */
class taka2Board {
public:
	piece_t t[TABLE_SIZE];
	cint_t mycount;
	cint_t matchcount;
	cint_t totalcount;
	cint_t step;		/* my steps */
	cint_t myflag;		/* True if this object is from my own view */
private:
	static int rev_table[4];
	static int boardmap_table[3];

public:
	taka2Board();
	taka2Board(taka2Board &src);
	taka2Board(char *initboard);
	int table(int pos);
	void settable(int pos, int d);
	void put_reverse(taka2Board &dstboard);	// for update
	taka2Board *create_reverse();
	
	// for debug use
#if TAKADEBUG
	void dump();
#endif
};

struct taka2Hand;
class taka2Choice;

class taka2Think {
	taka2Hand *handp;
public:	
	int weight;
protected:
	
	taka2Think(taka2Think &src);	// cannot call
	taka2Think();				// cannot call
	taka2Think(taka2Hand *handp);
	taka2Hand &hand();
	taka2Board &board();
	int piece_mpl(int color);
public:
	~taka2Think();
	
	// standard interfaces
	//taka2Think(taka2Hand *handp);
	//static void initialize();
	//static int think(taka2Board &bd, taka2Choice &chfirst);
};

/*
 * Thinking routine #1: 1997/09/26(Fri.)
 */
class taka2Think1 : public taka2Think {
public:
	// standard interfaces
	taka2Think1(taka2Hand *handp);
	~taka2Think1();
	static void initialize();
	static int think(taka2Board &bd, taka2Choice &chfirst);
	
private:
	// local data and functions
	taka2Choice *cp;
	/* taka2Board *bp; */
	int aheadresult;
	int lasthands;
	int evalflag;
	static int weight_table_seed[4][4];
	static int weight_table_seed2[4][4];
	static int weight_table[TABLE_SIZE];
	static int intable_quater[8][16];
	static int init_totalcount;
	static int total_breadth;
	
	void eval();
	void alleval();
	int finalize(int passcount, int orgmin);
	int prefinalize1(int passcount, int orgmin);
	int prefinalize1b(int passcount, int orgmin);
	int prefinalize2(int passcount, int orgmin, int limit);
	int semifinalize(int passcount, int orgmin);
	int nthforward(int passcount, int orgmin, int depth);
};
/* ----- */

struct taka2Hand {
	enum { maxhand=22, passed=-1 };
	taka2Board bd;
	getlist_t getlist[maxhand];
	cint_t pos;
	cint_t getcount;
private:
	// variables for thinking...
	Thinker th;
	
public:
	void *operator new(size_t siz);
	void operator delete(void *p);
	taka2Hand();
	~taka2Hand();
	Thinker &think();
	
	// for debug use
	//void dump();
};

class taka2Choice {
	enum { maxhand=60 };
	taka2Hand *hv[maxhand];
	int hc;
	int hrest;
	static taka2Hand *handstock;

protected:
	taka2Choice(taka2Choice &src);	// cannot call
	taka2Choice();					// cannot call
public:
	void *operator new(size_t siz);
	void operator delete(void *p);
	~taka2Choice();
	taka2Choice(taka2Board &from);
	void erasehand(int handcount);
	int number();
	int restnumber();
	taka2Hand &item(int hnum);	// const taka2Hand &
private:
	int makeput(taka2Hand *handp, taka2Board &from, int pos);
public:
	void handsort(int usecount = INT_MAX);
	// for debug use
	//void dump();
};


/*-----*/
inline void *taka2Hand::operator new(size_t /*siz*/)
{ return Malloc_block(&Malloc256); }
inline void taka2Hand::operator delete(void *p)
{ Free_block(&Malloc256, p); }
inline void *taka2Choice::operator new(size_t /*siz*/)
{ return Malloc_block(&Malloc256); }
inline void taka2Choice::operator delete(void *p)
{ Free_block(&Malloc256, p); }


/*----------------------------------------------------------------------*/
/*
 * tables
 */
STATIC int intable_table[8*8];

STATIC const int moves_table[8] = {
	-9-1, -9, -9+1,
	  -1,       +1,
	 9-1,  9,  9+1,
};
const int moves_table_0 = -9-1;
const int moves_table_1 = -9  ;
const int moves_table_2 = -9+1;
const int moves_table_3 =   -1;
const int moves_table_4 =   +1;
const int moves_table_5 =  9-1;
const int moves_table_6 =  9  ;
const int moves_table_7 =  9+1;


/*
 * global variables
 */
STATIC int my_step;

 /* established tactics table */
STATIC int estable_myc, *estable_my, *estable_myn;
STATIC int estable_opc, *estable_ops, *estable_opo, *estable_opn;
STATIC int estable_nowstat;
STATIC int estable_flipstate, estable_flip[4][TABLE_SIZE];
STATIC int estable_lastboard[8*8];


/*
 * prototype declarations
 */
STATIC void initialize_tables(void);
STATIC int estable_mystep();
STATIC void estable_opstep(int hand);
STATIC void estable_opstep(char *board);



/*----------------------------------------------------------------------*/
/*
 * class functions
 */

// class taka2Board
/* int taka2Board::rev_table[4] = { -1, 0, 2, 1 }; */
int taka2Board::rev_table[4] = { P_MY_COLOR, P_SPACE, P_OP_COLOR, P_WALL };
int taka2Board::boardmap_table[3] = { P_SPACE, P_MY_COLOR, P_OP_COLOR };

inline
taka2Board::taka2Board()
{
	;
}

inline
taka2Board::taka2Board(taka2Board &src) :
	mycount(src.mycount), matchcount(src.matchcount), totalcount(totalcount),
	step(src.step), myflag(src.myflag)
{
	memcpy(t, src.t, sizeof(piece_t)*TABLE_SIZE);
}

taka2Board::taka2Board(char *initboard)
{
	int l, d;

	for (l=0; l<9; l++) {
		t[tbl(0, l)] =
		t[tbl(l, 0)] =
		t[tbl(l, 9)] = P_WALL;
	}
	t[tbl(9,9)] = P_WALL;
	mycount = matchcount = 0;
	totalcount = 0;
	step = my_step;
	myflag = 1;
	for (l=0; l<64; l++) {
		t[intable_table[l]] = d = boardmap_table[initboard[l]];
		if (d == P_MY_COLOR) mycount++;
		else if (d == P_OP_COLOR) matchcount++;
	}
	totalcount = mycount + matchcount;
}

inline int
taka2Board::table(int pos)
{
	return t[pos];
}

inline void
taka2Board::settable(int pos, int d)
{
	t[pos] = d;
}

void
taka2Board::put_reverse(taka2Board &dstboard)
{
	int l;

	dstboard.mycount = matchcount;
	dstboard.matchcount = mycount;
	dstboard.totalcount = mycount + matchcount;
	if (myflag) {
		dstboard.step = step;
		dstboard.myflag = 0;
	}else {
		dstboard.step = step+1;
		dstboard.myflag = 1;
	}
	for (l=0; l<TABLE_SIZE; l++) {
		dstboard.t[l] = rev_table[1+t[l]];
	}
}

inline taka2Board *
taka2Board::create_reverse()
{
	taka2Board *rp = new taka2Board();
	put_reverse(*rp);
	return rp;
}

#if TAKADEBUG
void
taka2Board::dump()
{
	int x, y;
	int my, match;

	cout <<"taka: taka2Board " <<this <<" -> "
	  <<"mycount=" <<(int)mycount
	  <<", matchcount=" <<(int)matchcount
	  <<", step=" <<(int)step
	  <<", myflag=" <<(int)myflag <<endl;
	
	my = match = 0;
	for (y=0; y<=9; y++) {
		cout <<"  " <<y <<":";
		for (x=0; x<9; x++) {
			cout <<setw(3) <<(int)t[y*9+x];
			if (t[y*9+x] == P_MY_COLOR) my++;
			else if (t[y*9+x] == P_OP_COLOR) match++;
		}
		if (y==9) cout <<" ... " << (int)t[y*9+x];
		cout <<endl;
	}
	
	if (my != mycount) cout <<"  mycount mismatch! " <<my <<endl;
	if (match != matchcount) cout <<"  matchcount mismatch! " <<match <<endl;
}
#endif


// class taka2Hand
inline
taka2Hand::taka2Hand() : th(this)
{
	//tp = new Thinker(this);
}

inline
taka2Hand::~taka2Hand()
{
	//if (tp) delete tp;
}

inline Thinker &
taka2Hand::think()
{
	return th;
}


// class taka2Choice
taka2Hand *taka2Choice::handstock = new taka2Hand();

taka2Choice::taka2Choice(taka2Choice &/*src*/)
{
	abort();
}
taka2Choice::taka2Choice() : hc(0)
{
	abort();
}

taka2Choice::~taka2Choice()
{
	int l;

	if (!hc) {
		/* If hc==0, hv[0] keeps valid object to destroy. */
		delete hv[0];
#if TAKADEBUG
		hv[0] = (taka2Hand *)DEADBEEF;
#endif
		return;
	}
	for (l=0; l<hc; l++) {
		delete hv[l];
#if TAKADEBUG
		hv[l] = (taka2Hand *)DEADBEEF;
#endif
	}
}

inline void
taka2Choice::erasehand(int handcount)
{
	delete hv[handcount];
	hv[handcount] = NULL;
	hrest--;
}

taka2Choice::taka2Choice(taka2Board &from)
{
	int p;
	int l;
	//taka2Hand *hp;

#if TAKADEBUG
	for (l=0; l<maxhand; l++) hv[l] = (taka2Hand *)DEADBEEF;
#endif
	hc = 0;
	//hp = new taka2Hand();
	for (l=0; l<64; l++) {
		p = intable_table[l];
		if (from.table(p) != P_SPACE) continue;
		if (
		  from.table(p+moves_table_0) != P_OP_COLOR &&
		  from.table(p+moves_table_1) != P_OP_COLOR &&
		  from.table(p+moves_table_2) != P_OP_COLOR &&
		  from.table(p+moves_table_3) != P_OP_COLOR &&
		  from.table(p+moves_table_4) != P_OP_COLOR &&
		  from.table(p+moves_table_5) != P_OP_COLOR &&
		  from.table(p+moves_table_6) != P_OP_COLOR &&
		  from.table(p+moves_table_7) != P_OP_COLOR
		) {
			continue;
		}
		if (makeput(handstock, from, p)) {
			hv[hc++] = handstock;
			handstock = new taka2Hand();
		}
	}
	if (hc == 0) {
		/* pass */
		from.put_reverse(handstock->bd);
		/* handstock->bd = from; */
		handstock->getcount = 0;
		handstock->pos = taka2Hand::passed;
		//handstock->tp = new Thinker(handstock);
		hv[0] = handstock;
		handstock = new taka2Hand();
		hrest = 1;
	}else {
		//delete hp;
		hrest = hc;
	}
}

inline int
taka2Choice::number() {
	return hc;
}

inline int
taka2Choice::restnumber() {
	return hrest;
}

inline taka2Hand &
taka2Choice::item(int hnum) {
	return *hv[hnum];
}

int
taka2Choice::makeput(taka2Hand *handp, taka2Board &from, int pos)
{
	int l, p, md;
	int mc, gc;
	getlist_t *gv;

#if TAKADEBUG*0
	cout <<"makeput: " <<pos <<"(" <<tblx(pos) <<"," <<tbly(pos) <<")" <<endl;
#endif
	gc = 0;
	for (mc=0; mc<8; mc++) {
		if (from.table(pos+moves_table[mc]) != P_OP_COLOR) continue;
		md = moves_table[mc];
		gv = handp->getlist + gc;
		*gv = p = pos+md;
		l=1;
		while (from.table(p+=md) == P_OP_COLOR) {
			*(gv + l++) = p;
		}
		if (from.table(p) != P_MY_COLOR) continue;
		gc += l;
	}
	if (gc == 0) {
		return 0;
	}
	
	// update board table
	from.put_reverse(handp->bd);
	//handp->bd = from;
	handp->bd.settable(pos, P_OP_COLOR);
	for (l=0; l<gc; l++) {
		handp->bd.settable(handp->getlist[l], P_OP_COLOR);
	}
	handp->bd.matchcount += 1+gc;
	handp->bd.mycount -= gc;
	handp->bd.totalcount++;
	handp->pos = pos;
	handp->getcount = gc;
	//handp->tp = new Thinker(handp);
	
	return 1;
}

void
taka2Choice::handsort(int usecount)
{
	taka2Hand *hp;

	switch (number()) {
	case 0:
	case 1:
		return;
	case 2:
		if (item(0).think().weight > item(1).think().weight) {
			hp=hv[0]; hv[0]=hv[1]; hv[1]=hp;
		}
		return;
	case 3:
		if (item(0).think().weight > item(1).think().weight) {
			if (item(0).think().weight > item(2).think().weight) {
				if (item(1).think().weight > item(2).think().weight) {
					// 2->1->0
					hp=hv[0]; hv[0]=hv[2]; hv[2]=hp;
				}else {
					// 1->2->0
					hp=hv[0]; hv[0]=hv[1]; hv[1]=hv[2]; hv[2]=hp;
				}
			}else {
				// 1->0->2
				hp=hv[0]; hv[0]=hv[1]; hv[1]=hp;
			}
		}else {
			if (item(1).think().weight > item(2).think().weight) {
				if (item(0).think().weight > item(2).think().weight) {
					// 2->0->1
					hp=hv[0]; hv[0]=hv[2]; hv[2]=hv[1]; hv[1]=hp;
				}else {
					// 0->2->1
					hp=hv[1]; hv[1]=hv[2]; hv[2]=hp;
				}
			}
		}
		return;
	case 4:
		if (item(0).think().weight > item(2).think().weight) {
			hp=hv[0]; hv[0]=hv[2]; hv[2]=hp;
		}
		if (item(1).think().weight > item(3).think().weight) {
			hp=hv[1]; hv[1]=hv[3]; hv[3]=hp;
		}
		if (item(0).think().weight > item(1).think().weight) {
			hp=hv[0]; hv[0]=hv[1]; hv[1]=hp;
		}
		if (item(2).think().weight > item(3).think().weight) {
			hp=hv[2]; hv[2]=hv[3]; hv[3]=hp;
		}
		if (item(1).think().weight > item(2).think().weight) {
			hp=hv[1]; hv[1]=hv[2]; hv[2]=hp;
		}
		return;
	}
	{
		int e, l, i, p, m;
		
		// selective sort
		e = 2;
		if (usecount > e) e = usecount;
		if (e > number()-1) e = number()-1;
		for (l=0; l<e; l++) {
			m = item(l).think().weight;
			p = l;
			for (i=l+1; i<hc; i++) {
				if (item(i).think().weight < m) {
					m = item(i).think().weight;
					p = i;
				}
			}
			if (l != p) {
				hp=hv[l]; hv[l]=hv[p]; hv[p]=hp;
			}
		}
	}
}


// class taka2Think
taka2Think::taka2Think(taka2Think &/*src*/)
{
	abort();
}

taka2Think::taka2Think()
{
	abort();
}

inline
taka2Think::taka2Think(taka2Hand *handp) : handp(handp)
{
#if TAKADEBUG
	weight = 0;
#endif
	;
}

inline
taka2Think::~taka2Think()
{
#if TAKADEBUG
		handp = (taka2Hand *)DEADBEEF;
#endif
	;
}

inline taka2Hand &
taka2Think::hand()
{
	return *handp;
}

inline taka2Board &
taka2Think::board()
{
	return handp->bd;
}

inline int
taka2Think::piece_mpl(int color)
{
#if 0	/* P_SPACE=0, P_MY_COLOR=1, P_OP_COLOR=2 */
	static int weight_mpl[3] = { 0, 1, -1 };
	return weight_mpl[color];
#endif
	/* P_SPACE=0, P_MY_COLOR=1, P_OP_COLOR=-1 */
	return color;
}



/*----------------------------------------------------------------------*/
/*
 * utility and initialization functions
 */
STATIC void
initialize_tables(void)
{
	int x, y;
	int l, i, i2, i3;
	FILE *fp;
	static const int buflen = 128;
	char buf[buflen];

	for (y=0; y<8; y++) for (x=0; x<8; x++) {
		intable_table[y*8 + x] = tbl(x+1, y+1);
	}
#if TAKADEBUG*0
	for (int l=0; l<64; l++) {
		if (l%8 == 0) cout <<endl;
		cout <<setw(3) <<intable_table[l];
	}
	cout <<endl;
#endif
#if TAKADEBUG
	cout <<"initialize_tables: sizeof... taka2Hand=" <<sizeof(taka2Hand)
	  <<", taka2Choice=" <<sizeof(taka2Choice)
	  <<", Thinker=" <<sizeof(Thinker) <<"." <<endl;
#endif
	
	if (sizeof(taka2Choice) > Malloc256.block_size) {
#if TAKADEBUG
		cout <<"taka: sizeof(taka2Choice)=" <<sizeof(taka2Choice) <<endl;
#endif
		abort();
	}
	if (sizeof(taka2Hand) > Malloc256.block_size) {
#if TAKADEBUG
		cout <<"taka: sizeof(taka2Hand)=" <<sizeof(taka2Hand) <<endl;
#endif
		abort();
	}
	
#if 0
	int l = Malloc256.block_size * Malloc256.allocated_blocks * 4;
	int *p = (int *)malloc(l);
	l /= sizeof(int);
	p += l;
	l /= 256;
	do { *(p-=256) = 0; }while (--l);
	free(p);
#endif
	
	/* read established tactics data */
	if (estable_myc == 0) {
		if ((fp = fopen(ESTABLE_FILENAME, "r")) == NULL) {
			fprintf(stderr, "taka.cc: cannot open data file \x22%s\x22.\n",
			  ESTABLE_FILENAME);
			abort();
		}
		
		 /* read my table */
		do {
			if (fgets(buf, buflen, fp) == NULL) goto readerror;
		}while (strncmp(buf, "#MY ", 4));
		i = strtol(buf+4, NULL, 10);
		if (i <= 0) goto readerror;
		estable_myc = i;
		estable_my = (int *)malloc(estable_myc * sizeof(int));
		estable_myn = (int *)malloc(estable_myc * sizeof(int));
		if (estable_my == NULL || estable_myn == NULL) goto readerror;
		for (l=0; l<estable_myc; l++) {
			do {
				if (fgets(buf, buflen, fp) == NULL) goto readerror;
			}while (!buf[0]);
			if (sscanf(buf, "%d%d%d", &i, &i2, &i3) < 3) goto readerror;
			if (i != l || i2 < tbl(1,1) || i2 > tbl(8,8) || i3 < 1)
				goto readerror;
			estable_my[l] = i2;
			estable_myn[l] = i3;
		}
		
		do {
			if (fgets(buf, buflen, fp) == NULL) goto readerror;
		}while (!buf[0]);
		
		 /* read op table */
		if (fgets(buf, buflen, fp) == NULL) goto readerror;
		if (strncmp(buf, "#OP ", 4)) goto readerror;
		i = strtol(buf+4, NULL, 10);
		if (i <= 0) goto readerror;
		estable_opc = i;
		estable_ops = (int *)malloc(estable_opc * sizeof(int));
		estable_opo = (int *)malloc(estable_opc * sizeof(int));
		estable_opn = (int *)malloc(estable_opc * sizeof(int));
		if (estable_ops == NULL || estable_opo == NULL || estable_opn == NULL)
			goto readerror;
		for (l=0; l<estable_opc; l++) {
			do {
				if (fgets(buf, buflen, fp) == NULL) goto readerror;
			}while (!buf[0]);
			if (sscanf(buf, "%d%d%d", &i, &i2, &i3) < 3) goto readerror;
			if (i < 0 || i2 < tbl(1,1) || i2 > tbl(8,8) || i3 < 1)
				goto readerror;
			estable_ops[l] = i;
			estable_opo[l] = i2;
			estable_opn[l] = i3;
		}
		
		fclose(fp);
		
		/* create flip table */
		for (y=1; y<8; y++) for (x=1; x<8; x++) {
			estable_flip[0][tbl(x  , y  )] = //(4,3) standard
			estable_flip[1][tbl(y  , x  )] = //(3,4) x-y symmetric
			estable_flip[2][tbl(9-x, 9-y)] = //(5,6) point-symmetric
			estable_flip[3][tbl(9-y, 9-x)] = //(6,5) mixture
			  tbl(x,y);
		}
		
	}
	
	estable_nowstat = 0;
	
	for (y=0; y<8; y++) for (x=0; x<8; x++) estable_lastboard[y*8+x] = 0;
	estable_lastboard[3*8+3] = estable_lastboard[3*8+4] = 
	estable_lastboard[4*8+3] = estable_lastboard[4*8+4] = 1;
		

	return;
	
readerror:
	fclose(fp);
	fprintf(stderr, "taka.cc: data file corrupted.\n");
	abort();	
}

STATIC int
estable_mystep()
{
	int r;

#if TAKADEBUG*0
	printf("estable_myc=%d, estable_nowstat=%d\n", estable_myc, estable_nowstat);
#endif
	if (estable_myc == 0) return -1;
	if (estable_nowstat < 0) return -1;
	if (estable_nowstat >= estable_myc) {
		estable_nowstat = -1;
		return -1;
	}
	if (estable_nowstat == 0) estable_flipstate = 0;
	
	r = estable_my[estable_nowstat];
#if TAKADEBUG
	cout <<"taka: estable my " <<estable_nowstat <<" -> "
	  <<tblx(r) <<"," <<tbly(r) <<" next " <<estable_myn[estable_nowstat]
	  <<"." <<endl;
#endif
	estable_nowstat = estable_myn[estable_nowstat];
	
	estable_lastboard[tblresult(estable_flip[estable_flipstate][r])] = 1;
	
	return estable_flip[estable_flipstate][r];
}

STATIC void
estable_opstep(int hand)
{
	int l;

	if (estable_myc == 0) return;
	if (estable_nowstat < 0) return;
	if (estable_nowstat == 0) {
		if (hand == estable_flip[0][estable_my[0]]) estable_flipstate = 0;
		else if (hand == estable_flip[1][estable_my[0]]) estable_flipstate = 1;
		else if (hand == estable_flip[2][estable_my[0]]) estable_flipstate = 2;
		else if (hand == estable_flip[3][estable_my[0]]) estable_flipstate = 3;
		else estable_nowstat = -1;
#if TAKADEBUG
		cout <<"taka: estable op init " <<estable_nowstat <<" flip "
		  <<estable_flipstate <<"." <<endl;
#endif
	}
	
	hand = estable_flip[estable_flipstate][hand];
	for (l=0; l<estable_opc; l++) {
		if (estable_ops[l] == estable_nowstat && estable_opo[l] == hand) {
#if TAKADEBUG
			cout <<"taka: estable op " <<estable_nowstat <<" -> "
			  <<estable_opn[l] <<"." <<endl;
			
#endif
			estable_nowstat = estable_opn[l];
			return;
		}
	}
#if TAKADEBUG
	cout <<"taka: estable op " <<estable_nowstat <<" -> " <<hand
	  <<" not found." <<endl;
			
#endif
	estable_nowstat = -1;
}

STATIC void
estable_opstep(char *board)
{
	int x,y;

#if TAKADEBUG*0
	if (my_step < 1) {
		for (y=0; y<8; y++) {
			for (x=0; x<8; x++) printf(" %d", (int)board[y*8+x]);
			printf("\n");
		}
	}
#endif
	if (estable_myc == 0) return;
	if (estable_nowstat < 0) return;
	
	for (y=0; y<8; y++) for (x=0; x<8; x++) {
		if (board[y*8+x] != SPACE && !estable_lastboard[y*8+x]) {
#if TAKADEBUG
			cout <<"taka: estable op hand = " <<x <<"," <<y <<endl;
#endif
			estable_opstep(tbl(x+1,y+1));
			estable_lastboard[y*8+x] = 1;
			return;
		}
	}
	// ꤬ѥ
}


/*
 * entry of thinking function
 */
extern "C" int
taka2(char *board, int status)
{
	int p;

	if (status) {
		if (initp(status)) {
			/* initialization for static variables */
			my_step = 0;
			initialize_tables();
			Thinker::initialize();
			
			return 0;
		}
		if (passp(status)) {
			/* pass! -  check match's hand */
			estable_opstep(board);
			my_step++;
			
			return 0;
		}
#if TAKADEBUG
		cout <<"taka .. unknown status." <<endl;
#endif
		abort();
	}
	
	estable_opstep(board);
	
	my_step++;
	
	p = estable_mystep();
	if (p > 0) {
#if TAKADEBUG
		cout <<"taka: estable hit. " <<endl;
#endif
		return tblresult(p);
	}
	
	taka2Board bd(board);
	taka2Choice chfirst(bd);
	if (chfirst.number() <= 1) {
		if (chfirst.number() == 0) {
			/* pass */
			//return -1;
#if TAKADEBUG
			cout <<"taka .. unexpected pass situation." <<endl;
#endif
			abort();
		}
		/* no room to choose */
		p = chfirst.item(0).pos;
		goto thank;
	}
	
	p = Thinker::think(bd, chfirst);

thank:
	return tblresult(p);
}


/*----------------------------------------------------------------------*/

/*
 * Thinking routine #1: 1997/09/26(Fri.)
 */
// static variables
int taka2Think1::weight_table_seed[4][4] = {
	{  990,   10,  140,   90, },
	{   10,   60,  120,   80, },
	{  140,  120,   50,   60, },
	{   90,   80,   60,   50, },
};
int taka2Think1::weight_table_seed2[4][4] = {
	{  990,  -50,  280,  100, },
	{  -50,   60,  260,   80, },
	{  280,  260,   50,   60, },
	{  100,   80,   60,   60, },
};
int taka2Think1::weight_table[TABLE_SIZE];
int taka2Think1::intable_quater[8][16];
int taka2Think1::init_totalcount;
int taka2Think1::total_breadth;

// functions
taka2Think1::taka2Think1(taka2Hand *handp) : taka2Think(handp)
{
	cp = NULL;
	aheadresult = INT_MAX;
	evalflag = 0;
	/* bp = NULL; */
}

taka2Think1::~taka2Think1()
{
	if (cp != NULL) {
		delete cp;
#if TAKADEBUG
		cp = (taka2Choice *)DEADBEEF;
#endif
	}
#if 0
	if (bp != NULL) {
		delete bp;
#if TAKADEBUG
		bp = (taka2Board *)DEADBEEF;
#endif
	}
#endif
}

void
taka2Think1::initialize()
{
	int l, x, y;

	for (l=0; l<=9; l++) {
		weight_table[tbl(0, l)] =
		weight_table[tbl(9, l)] =
		weight_table[tbl(l, 0)] =
		weight_table[tbl(l, 9)] = 0;
	}
	for (y=1; y<=4; y++) for (x=1; x<=4; x++) {
		weight_table[tbl(x,   y  )] =
		weight_table[tbl(9-x, y  )] =
		weight_table[tbl(x,   9-y)] =
		weight_table[tbl(9-x, 9-y)] = weight_table_seed2[y-1][x-1];
	}
	
	for (y=0; y<4; y++) for (x=0; x<4; x++) {
		intable_quater[0][y*4+x] =
		intable_quater[1][x*4+y] = tbl(1+x, 1+y);
		intable_quater[2][y*4+x] =
		intable_quater[3][x*4+y] = tbl(8-x, 1+y);
		intable_quater[4][y*4+x] =
		intable_quater[5][x*4+y] = tbl(1+x, 8-y);
		intable_quater[6][y*4+x] =
		intable_quater[7][x*4+y] = tbl(8-x, 8-y);
	}
	
	
#if TAKADEBUG*0
	for (l=0; l<91; l++) {
		if (l%9 == 0) cout <<endl;
		cout <<setw(3) <<weight_table[l];
	}
	cout <<endl;
#endif
}

void
taka2Think1::eval()
{
	int l;
	int v;
	int *itqp, *wtqp;
	/*
	int d;
	int x,y;
	*/
	int mc, oc;

	static int weight_table_a1[4][4] = {
		{ 1500,   20,  100,   80, },
		{   20,   60,   70,   80, },
		{  100,   70,   50,   60, },
		{   80,   80,   60,   50, },
	};
	static int weight_table_a2[4][4] = {
		{ 1500, -280,   80,   90, },
		{ -200,   80,  100,   70, },
		{  200,  200,   50,   60, },
		{  110,  100,   60,   50, },
	};
	static int weight_table_a3[4][4] = {
		{ 1500, -200,   90,   90, },
		{ -200,   60,  100,   80, },
		{   90,  100,   50,   60, },
		{  100,   80,   60,   50, },
	};
/*	
	{ 150,   1,  14,   9, },
	{   1,   6,  12,   8, },
	{  14,  12,   5,   6, },
	{   9,   8,   6,   5, },
*/
	static int weight_table_b2[4][4] = {
		{  990,   10,  100,  100, },
		{ -200,   80,   90,   70, },
		{  180,  160,   50,   70, },
		{   90,   80,   70,   60, },
	};
	static int weight_table_b3[4][4] = {
		{  990,   10,  100,   90, },
		{   10,   50,   80,   70, },
		{  100,   80,   50,   60, },
		{   90,   70,   60,   50, },
	};
	static int narrow_penalty[7] = {
		200, 700, 500, 300, 200, 90, 5
	};

	if (board().totalcount <= /*38*/60) {
		v = 0;
		
		for (l=0; l<8; l+=2) {
			itqp = intable_quater[l];
			if (board().table(itqp[0]) != P_SPACE) {
				wtqp = (int *)weight_table_a1;
			}else if (board().table(itqp[1]) != P_SPACE) {
				if (board().table(itqp[4]) != P_SPACE) {
					/*wtqp = (board().totalcount <= 34)?*/
					wtqp = (init_totalcount <= 34)?
					  (int *)weight_table_a3 : (int *)weight_table_b3;
				}else {
					wtqp = (init_totalcount <= 34)?
					  (int *)weight_table_a2 : (int *)weight_table_b2;
				}
			}else if (board().table(itqp[4]) != P_SPACE) {
				itqp = intable_quater[l+1];  /* swap x and y */
				wtqp = (init_totalcount <= 34)?
				  (int *)weight_table_a2 : (int *)weight_table_b2;
			}else {
#if 0
				for (x=0; x<16; x++) {
					d = board().table(itqp[x]);
					if (d == P_MY_COLOR) {
						v += weight_table[itqp[x]];
					}else if (d == P_OP_COLOR) {
						v -= weight_table[itqp[x]];
					}
				}
#endif
				int s[4+2], *sp;
				piece_t *t;
				
				/* use standard table */
				sp = s+1+2; sp[P_OP_COLOR] = sp[P_MY_COLOR] = 0;
				t = board().t;
				
				sp[(int)t[itqp[0]]] += weight_table[itqp[0]];
				sp[(int)t[itqp[1]]] += weight_table[itqp[1]];
				sp[(int)t[itqp[2]]] += weight_table[itqp[2]];
				sp[(int)t[itqp[3]]] += weight_table[itqp[3]];
				sp[(int)t[itqp[4]]] += weight_table[itqp[4]];
				sp[(int)t[itqp[5]]] += weight_table[itqp[5]];
				sp[(int)t[itqp[6]]] += weight_table[itqp[6]];
				sp[(int)t[itqp[7]]] += weight_table[itqp[7]];
				sp[(int)t[itqp[8]]] += weight_table[itqp[8]];
				sp[(int)t[itqp[9]]] += weight_table[itqp[9]];
				sp[(int)t[itqp[10]]] += weight_table[itqp[10]];
				sp[(int)t[itqp[11]]] += weight_table[itqp[11]];
				sp[(int)t[itqp[12]]] += weight_table[itqp[12]];
				sp[(int)t[itqp[13]]] += weight_table[itqp[13]];
				sp[(int)t[itqp[14]]] += weight_table[itqp[14]];
				sp[(int)t[itqp[15]]] += weight_table[itqp[15]];
				
				v += sp[P_MY_COLOR] - sp[P_OP_COLOR];
				
				continue;
			}
			
#if 0
			for (x=0; x<16; x++) {
				d = board().table(itqp[x]);
				if (d == P_MY_COLOR) {
					v += wtqp[x];
				}else if (d == P_OP_COLOR) {
					v -= wtqp[x];
				}
			}
#endif
			{
			int s[4+1], *sp;
			piece_t *t;
			
			sp = s+1+1; sp[P_OP_COLOR] = sp[P_MY_COLOR] = 0;
			t = board().t;
			
			sp[(int)t[itqp[0]]] += wtqp[0];
			sp[(int)t[itqp[1]]] += wtqp[1];
			sp[(int)t[itqp[2]]] += wtqp[2];
			sp[(int)t[itqp[3]]] += wtqp[3];
			sp[(int)t[itqp[4]]] += wtqp[4];
			sp[(int)t[itqp[5]]] += wtqp[5];
			sp[(int)t[itqp[6]]] += wtqp[6];
			sp[(int)t[itqp[7]]] += wtqp[7];
			sp[(int)t[itqp[8]]] += wtqp[8];
			sp[(int)t[itqp[9]]] += wtqp[9];
			sp[(int)t[itqp[10]]] += wtqp[10];
			sp[(int)t[itqp[11]]] += wtqp[11];
			sp[(int)t[itqp[12]]] += wtqp[12];
			sp[(int)t[itqp[13]]] += wtqp[13];
			sp[(int)t[itqp[14]]] += wtqp[14];
			sp[(int)t[itqp[15]]] += wtqp[15];
			
			v += sp[P_MY_COLOR] - sp[P_OP_COLOR];
			}
		}
		
		 /* board() has the state in complementary color */
		weight = - v;
		
		if (board().totalcount >= 58 || board().totalcount <= 12 ||
		  (init_totalcount == 50 || init_totalcount == 51)
		    && board().totalcount == 55+1
		) return;
		
		if (cp == NULL) {
			cp = new taka2Choice(board());
			for (l=0; l<cp->number(); l++) {
				cp->item(l).think().lasthands = cp->number();
			}
		}
		
		mc = cp->number();
		oc = lasthands;
		
		if (mc <= 6) weight -= narrow_penalty[mc];
		if (oc <= 6) weight += narrow_penalty[oc];
		
		mc = (mc>16) ? mc+16 : mc*2;
		oc = (oc>16) ? oc+16 : oc*2;
		
		if (board().totalcount < 24) {
			weight -= (oc-mc)*(127/2);
		}else if (board().totalcount < 38) {
			weight -= (oc-mc)*(97/2);
		}else if (board().totalcount < 52) {
			weight -= (oc-mc)*(73/2);
		}else {
			weight -= (oc-mc)*(59/2);
		}
		
		return;
	}
	
	if (board().totalcount < 60) {
#if 0
		v = 0;
		for (l=0; l<64; l++) {
			d = board().table(intable_table[l]);
			if (d == P_MY_COLOR) {
				v += weight_table[intable_table[l]];
			}else if (d == P_OP_COLOR) {
				v -= weight_table[intable_table[l]];
			}
		}
		 /* board() has the state in complementary color */
		weight = - v;
#endif
		int s[4], *sp;
		piece_t *t;
		int w,c1,c2;
		
		sp = s+1; sp[P_OP_COLOR] = sp[P_MY_COLOR] = 0;
		t = board().t;
		
		c1 = t[tbl(1,1)];              w = weight_table[tbl(1,1)];
		c2 = t[tbl(8,1)]; sp[c1] += w;
		c1 = t[tbl(1,8)]; sp[c2] += w;
		c2 = t[tbl(8,8)]; sp[c1] += w;
		c1 = t[tbl(2,1)]; sp[c2] += w; w = weight_table[tbl(2,1)];
		c2 = t[tbl(7,1)]; sp[c1] += w;
		c1 = t[tbl(2,8)]; sp[c2] += w;
		c2 = t[tbl(7,8)]; sp[c1] += w;
		c1 = t[tbl(3,1)]; sp[c2] += w; w = weight_table[tbl(3,1)];
		c2 = t[tbl(6,1)]; sp[c1] += w;
		c1 = t[tbl(3,8)]; sp[c2] += w;
		c2 = t[tbl(6,8)]; sp[c1] += w;
		c1 = t[tbl(4,1)]; sp[c2] += w; w = weight_table[tbl(4,1)];
		c2 = t[tbl(5,1)]; sp[c1] += w;
		c1 = t[tbl(4,8)]; sp[c2] += w;
		c2 = t[tbl(5,8)]; sp[c1] += w;
		
		c1 = t[tbl(1,2)]; sp[c2] += w; w = weight_table[tbl(1,2)];
		c2 = t[tbl(8,2)]; sp[c1] += w;
		c1 = t[tbl(1,7)]; sp[c2] += w;
		c2 = t[tbl(8,7)]; sp[c1] += w;
		c1 = t[tbl(2,2)]; sp[c2] += w; w = weight_table[tbl(2,2)];
		c2 = t[tbl(7,2)]; sp[c1] += w;
		c1 = t[tbl(2,7)]; sp[c2] += w;
		c2 = t[tbl(7,7)]; sp[c1] += w;
		c1 = t[tbl(3,2)]; sp[c2] += w; w = weight_table[tbl(3,2)];
		c2 = t[tbl(6,2)]; sp[c1] += w;
		c1 = t[tbl(3,7)]; sp[c2] += w;
		c2 = t[tbl(6,7)]; sp[c1] += w;
		c1 = t[tbl(4,2)]; sp[c2] += w; w = weight_table[tbl(4,2)];
		c2 = t[tbl(5,2)]; sp[c1] += w;
		c1 = t[tbl(4,7)]; sp[c2] += w;
		c2 = t[tbl(5,7)]; sp[c1] += w;
		
		c1 = t[tbl(1,3)]; sp[c2] += w; w = weight_table[tbl(1,3)];
		c2 = t[tbl(8,3)]; sp[c1] += w;
		c1 = t[tbl(1,6)]; sp[c2] += w;
		c2 = t[tbl(8,6)]; sp[c1] += w;
		c1 = t[tbl(2,3)]; sp[c2] += w; w = weight_table[tbl(2,3)];
		c2 = t[tbl(7,3)]; sp[c1] += w;
		c1 = t[tbl(2,6)]; sp[c2] += w;
		c2 = t[tbl(7,6)]; sp[c1] += w;
		c1 = t[tbl(3,3)]; sp[c2] += w; w = weight_table[tbl(3,3)];
		c2 = t[tbl(6,3)]; sp[c1] += w;
		c1 = t[tbl(3,6)]; sp[c2] += w;
		c2 = t[tbl(6,6)]; sp[c1] += w;
		c1 = t[tbl(4,3)]; sp[c2] += w; w = weight_table[tbl(4,3)];
		c2 = t[tbl(5,3)]; sp[c1] += w;
		c1 = t[tbl(4,6)]; sp[c2] += w;
		c2 = t[tbl(5,6)]; sp[c1] += w;
		
		c1 = t[tbl(1,4)]; sp[c2] += w; w = weight_table[tbl(1,4)];
		c2 = t[tbl(8,4)]; sp[c1] += w;
		c1 = t[tbl(1,5)]; sp[c2] += w;
		c2 = t[tbl(8,5)]; sp[c1] += w;
		c1 = t[tbl(2,4)]; sp[c2] += w; w = weight_table[tbl(2,4)];
		c2 = t[tbl(7,4)]; sp[c1] += w;
		c1 = t[tbl(2,5)]; sp[c2] += w;
		c2 = t[tbl(7,5)]; sp[c1] += w;
		c1 = t[tbl(3,4)]; sp[c2] += w; w = weight_table[tbl(3,4)];
		c2 = t[tbl(6,4)]; sp[c1] += w;
		c1 = t[tbl(3,5)]; sp[c2] += w;
		c2 = t[tbl(6,5)]; sp[c1] += w;
		c1 = t[tbl(4,4)]; sp[c2] += w; w = weight_table[tbl(4,4)];
		c2 = t[tbl(5,4)]; sp[c1] += w;
		c1 = t[tbl(4,5)]; sp[c2] += w;
		c2 = t[tbl(5,5)]; sp[c1] += w;
		
		                  sp[c2] += w;
		
		 /* board() has the state in complementary color */
		weight = -(sp[P_MY_COLOR] - sp[P_OP_COLOR]);
		
		return;
	}
	
	 /* board() has the state in complementary color */
	weight = - (board().mycount - board().matchcount);
	
	return;
}

void
taka2Think1::alleval()
{
	int l;

	if (evalflag) return;
	if (cp->number() == 0) {
		cp->item(0).think().eval();
	}else {
		for (l=0; l<cp->number(); l++) cp->item(l).think().eval();
	}
	evalflag = 1;
}

int
taka2Think1::think(taka2Board &bd, taka2Choice &chfirst)
{
	int l;
	int c;
	int e;
	int lv;
	int max, min, maxpos, minpos;
	int n, nm;
	int keepminpospos;

	init_totalcount = bd.totalcount;
	total_breadth = 0;
	for (l=0; l<chfirst.number(); l++) {
		chfirst.item(l).think().lasthands = chfirst.number();
	}
	
	if (bd.totalcount >= 53) {
#if TAKADEBUG
		cout <<"taka: " <<(64 - bd.mycount - bd.matchcount)
		  <<"ꡢɤߤޤ(" <<chfirst.number() <<"̤) "<<flush;
#endif
		for (l=0; l<chfirst.number(); l++) {
			chfirst.item(l).think().eval();
		}
		chfirst.handsort();
		
		min = INT_MAX; minpos = -1;
		for (l=0; l<chfirst.number(); l++) {
#if TAKADEBUG
			cout <<l <<flush;
#endif
			c = - chfirst.item(l).think().finalize(0, min);
#if TAKADEBUG
			cout <<"<" <<c <<"> "<<flush;
#endif
			if (c < min) {
				if (minpos >= 0) chfirst.erasehand(minpos);
				min = c;
				minpos = l;
			}else {
				chfirst.erasehand(l);
			}
		}
#if TAKADEBUG
		cout <<" -> " <<minpos <<"<" <<min <<">" <<endl;
#endif
		return chfirst.item(minpos).pos;
	}
	
	
	if (bd.totalcount >= 50) {
#if TAKADEBUG
		cout <<"taka: " <<(64 - bd.mycount - bd.matchcount)
		  <<"ꡢԴɤߤޤ(" <<chfirst.number() <<"̤) "<<flush;
#endif
		for (l=0; l<chfirst.number(); l++) {
			chfirst.item(l).think().eval();
		}
		chfirst.handsort();
		
		min = INT_MAX; minpos = -1;
		for (l=0; l<chfirst.number(); l++) {
#if TAKADEBUG
			cout <<l <<flush;
#endif
			if (bd.totalcount <= 50) {
				chfirst.item(l).think().aheadresult
				  = c = - chfirst.item(l).think().prefinalize1b(0, min);
			}else {
				chfirst.item(l).think().aheadresult
				  = c = - chfirst.item(l).think().prefinalize1(0, min);
			}
#if TAKADEBUG
			//cout <<"<" <<c <<"> "<<flush;
			cout <<"<" <<c <<":" <<chfirst.item(l).think().cp->number()
			  <<">" <<total_breadth <<" "<<flush;
#endif
			if (c < min) {
				//if (minpos >= 0) chfirst.erasehand(minpos);
				if (c <= 0 && min > 0) {
					int i;
					for (i=0; i<l; i++) chfirst.erasehand(i);
				}
				min = c;
				minpos = l;
			}else {
				//chfirst.erasehand(l);
				if (min <= 0 && c > 0) chfirst.erasehand(l);
			}
			if (min<0 && l>=4) { l++; break; }
			if (bd.totalcount <= 50 && total_breadth > 500) break;
		}
#if TAKADEBUG
		cout <<" -> " <<minpos <<"<" <<min <<">" <<endl;
#endif
		keepminpospos = chfirst.item(minpos).pos;
		/*
		if (bd.totalcount <= 50) {
			return chfirst.item(minpos).pos;
		}
		*/
		
		nm = l;
		c = 0;
		if (min < 0) {
			for (l=0; l<nm; l++) {
				if (&chfirst.item(l) != NULL) {
					if (chfirst.item(l).think().aheadresult >= 0) {
						chfirst.erasehand(l);
					}else {
						c++;
					}
				}
			}
			if (bd.totalcount <= 50) min = INT_MAX;
		}else {
			min = INT_MAX;
		}
		
		if (c == 1) return chfirst.item(minpos).pos;

#if TAKADEBUG
		cout <<"        ɤߤޤ " <<flush;
#endif
		minpos = -1;
		n = 0;
		for (l=0; l<nm; l++) {
#if TAKADEBUG
			cout <<l <<flush;
#endif
			if (&chfirst.item(l) == NULL) continue;
			c = - chfirst.item(l).think().prefinalize2(0, min, -min);
#if TAKADEBUG
			//cout <<"<" <<c <<"> "<<flush;
			cout <<"<" <<c <<":" <<chfirst.item(l).think().cp->number()
			  <<">" <<total_breadth <<" "<<flush;
#endif
			if (c <= min && (c<min || minpos<0)) {
				if (minpos >= 0) chfirst.erasehand(minpos);
				min = c;
				minpos = l;
			}else {
				chfirst.erasehand(l);
			}
			if (++n>=4) break;
			if (bd.totalcount <= 50 && total_breadth > 1300) break;
		}
#if TAKADEBUG
		cout <<" -> " <<minpos <<"<" <<min <<">" <<endl;
#endif
		
		if (minpos < 0) return keepminpospos;
		return chfirst.item(minpos).pos;
	}
	
	
	if (bd.totalcount >= 48) {
#if TAKADEBUG
		cout <<"taka: " <<(64 - bd.mycount - bd.matchcount)
		  <<"ꡢɤߤޤ(" <<chfirst.number() <<"̤) "<<flush;
#endif
		for (l=0; l<chfirst.number(); l++) {
			chfirst.item(l).think().eval();
		}
		chfirst.handsort();
#if TAKADEBUG*0
		cout <<"sorted.. ";
		for (l=0; l<chfirst.number(); l++) {
			cout <<chfirst.item(l).think().weight <<" ";
		}
		cout <<endl;
#endif
		min = INT_MAX; minpos = -1;
		e = chfirst.number();
		if (e>4) e=4;
		for (l=0; l<e; l++) {
#if TAKADEBUG
			cout <<l <<flush;
#endif
			c = - chfirst.item(l).think().semifinalize(0, min);
#if TAKADEBUG
			//cout <<"<" <<c <<"> "<<flush;
			cout <<"<" <<c <<":" <<chfirst.item(l).think().cp->number()
			  <<">" <<total_breadth <<" "<<flush;
#endif
			if (c < min) {
				if (minpos >= 0) chfirst.erasehand(minpos);
				min = c;
				minpos = l;
			}else {
				chfirst.erasehand(l);
			}
			if (min < -2) {
				if (l>=3) break;
				if (l>=1 && bd.totalcount <= 48) break;
			}else {
				if (l>=5) break;
				if (l>=3 && bd.totalcount <= 48) break;
			}
			if (bd.totalcount <= 48 && total_breadth > 13000) break;
		}
#if TAKADEBUG
		cout <<" -> " <<minpos <<"<" <<min <<">" <<endl;
#endif
		return chfirst.item(minpos).pos;
	}
	
	
#if 0
	if (bd.totalcount >= 9) {
		lv = 3-1;
	if (bd.totalcount >= 14) {
		lv = 5-1;
		if (bd.totalcount >= 28) {
			lv = 6-1;
			if (bd.totalcount >= 38) {
				lv = 7-1;
				if (bd.totalcount >= 44) {
					lv = 8-1;
				if (bd.totalcount >= 47) {
					lv = 9-1;
#endif
	if (bd.totalcount >= 6) {
		lv = 3-1;
	if (bd.totalcount >= 10) {
		lv = 4-1;
		if (bd.totalcount >= 15) {
			lv = 6-1;
			if (bd.totalcount >= 30) {
				lv = 8-1;
				if (bd.totalcount >= 44) {
					lv = 8-1;
				if (bd.totalcount >= 47) {
					lv = 9-1;
				}
				}
			}
		}
		}
#if TAKADEBUG
		cout <<"taka: " <<(64 - bd.mycount - bd.matchcount) <<"ꡢ"
		  <<lv+1 <<"ɤߤޤ(" <<chfirst.number() <<"̤)" <<flush;
#endif
		for (l=0; l<chfirst.number(); l++) {
			chfirst.item(l).think().eval();
		}
		chfirst.handsort();
#if TAKADEBUG*1
		cout <<"sorted.. ";
		for (l=0; l<chfirst.number(); l++) {
			cout <<chfirst.item(l).think().weight <<" ";
		}
		cout <<endl;
#endif
		if (bd.totalcount == 26 || bd.totalcount >= 16 && bd.totalcount < 26 && chfirst.item(0).think().weight < -10) {
			int x,y;
			for (y=1; y<=4; y++) for (x=1; x<=4; x++) {
				weight_table[tbl(x,   y  )] =
				weight_table[tbl(9-x, y  )] =
				weight_table[tbl(x,   9-y)] =
				weight_table[tbl(9-x, 9-y)] = weight_table_seed[y-1][x-1];
			}
		}
	
		min = INT_MAX; minpos = -1;
		e = chfirst.number();
		if (e>4) {
			if (bd.totalcount >= 30) e=4;
			else if (e>6) e=6;
		}
		for (l=0; l<e; l++) {
#if TAKADEBUG
			cout <<l <<flush;
#endif
			c = - chfirst.item(l).think().nthforward(0, min, lv);
#if TAKADEBUG
			cout <<"<" <<c <<":" <<chfirst.item(l).think().cp->number()
			  <<"> "<<flush;
#endif
			if (c < min) {
				if (minpos >= 0) chfirst.erasehand(minpos);
				min = c;
				minpos = l;
			}else {
				chfirst.erasehand(l);
			}
			if (min<0) {
				if (l>=1 && bd.totalcount >= 48) break;
				if (l>=2 && bd.totalcount >= 33) break;
			}
		}
#if TAKADEBUG
		cout <<" -> " <<minpos <<"<" <<min <<">" <<endl;
#endif
		return chfirst.item(minpos).pos;
	}
	
	
	for (l=0; l<chfirst.number(); l++) {
		chfirst.item(l).think().eval();
	}
	max = -INT_MAX; min = INT_MAX; maxpos=minpos=0;
	for (l=0; l<chfirst.number(); l++) {
		if (chfirst.item(l).think().weight > max) {
			max = chfirst.item(l).think().weight;
			maxpos = l;
		}
		if (chfirst.item(l).think().weight < min) {
			min = chfirst.item(l).think().weight;
			minpos = l;
		}
	}
	
	return chfirst.item(minpos).pos;
}

int
taka2Think1::finalize(int passcount, int orgmin)
{
	int l;
	int c, min, minpos;

	/*
	if (board().totalcount == 64) {
		return (board().mycount - board().matchcount);
	}
	*/
	/* bp = board().create_reverse(); */
	if (cp == NULL) {
		cp = new taka2Choice(board());
		for (l=0; l<cp->number(); l++) {
			cp->item(l).think().lasthands = cp->number();
		}
	}
	if (cp->number() <= 1) {
		if (cp->number() == 0) {
			/* pass */
			if (passcount) {
				/* both pass */
				return (board().mycount - board().matchcount);
			}
			return - cp->item(0).think().finalize(1, INT_MAX);
		}
		/* no room to choose */
		if (board().totalcount == 63) {
			return -(cp->item(0).bd.mycount - cp->item(0).bd.matchcount);
		}
		return - cp->item(0).think().finalize(0, INT_MAX);
	}
	
	if (orgmin < INT_MAX) {
		alleval();
		/* for (l=0; l<cp->number(); l++) cp->item(l).think().eval(); */
		cp->handsort(8);
	}
	
	min = INT_MAX; minpos = -1; c = 0;
	for (l=0; l<cp->number(); l++) {
		if (board().totalcount == 63) {
			c = -(cp->item(l).bd.mycount - cp->item(l).bd.matchcount);
		}else {
			c = - cp->item(l).think().finalize(0, min);
		}
		if (c < min) {
			if (minpos >= 0) cp->erasehand(minpos);
			min = c;
			minpos = l;
		}else {
			cp->erasehand(l);
		}
		if (orgmin <= -min) break;	// alpha-beta
	}

	return min;
}

int
taka2Think1::prefinalize1(int passcount, int orgmin)
{
	int l;
	int c, min, minpos;

	if (board().totalcount == 64) {
		return (board().mycount - board().matchcount);
	}
	if (cp == NULL) {
		cp = new taka2Choice(board());
		for (l=0; l<cp->number(); l++) {
			cp->item(l).think().lasthands = cp->number();
		}
	}
	if (cp->number() <= 1) {
		if (cp->number() == 0) {
			/* pass */
			if (passcount) {
				/* both pass */
				return (board().mycount - board().matchcount);
			}
			return - cp->item(0).think().prefinalize1(1, INT_MAX);
		}
		/* no room to choose */
		if (board().totalcount<=55) {
			cp->item(0).think().aheadresult
			  = c = - cp->item(0).think().prefinalize1(0, INT_MAX);
		}else {
			cp->item(0).think().aheadresult
			  = c = - cp->item(0).think().finalize(0, INT_MAX);
		}
		return c;
	}
	
	alleval();
	/* for (l=0; l<cp->number(); l++) cp->item(l).think().eval(); */
	cp->handsort(6);
	
	/*
	if (board().totalcount<=55) {
		for (l=0; l<cp->number(); l++)
			cp->item(l).think().aheadresult = INT_MAX;
	}
	*/
	
	min = INT_MAX; minpos = -1; c = 0;
	for (l=0; l<cp->number(); l++) {
		if (board().totalcount<=55) {
			cp->item(l).think().aheadresult
			  = c = - cp->item(l).think().prefinalize1(0, min);
		}else {
			cp->item(l).think().aheadresult
			  = c = - cp->item(l).think().finalize(0, min);
		}
		if (c < min) {
			//if (minpos >= 0) cp->erasehand(minpos);
			min = c;
			minpos = l;
		}else {
			//cp->erasehand(l);
		}
		if (orgmin <= -min) { l++; break; }	// alpha-beta
		if (min<0 /*&& l>=1*/) { l++; break; }
	}
	
	if (min < 0) {
		for (c=0; c<l; c++) {
			if (cp->item(c).think().aheadresult >= 0) cp->erasehand(c);
		}
	}

	return min;
}

int
taka2Think1::prefinalize1b(int passcount, int orgmin)
{
	int l;
	int c, min, minpos;

	if (board().totalcount == 64) {
		return (board().mycount - board().matchcount);
	}
	total_breadth++;
	
	if (cp == NULL) {
		cp = new taka2Choice(board());
		for (l=0; l<cp->number(); l++) {
			cp->item(l).think().lasthands = cp->number();
		}
	}
	if (cp->number() <= 1) {
		if (cp->number() == 0) {
			/* pass */
			if (passcount) {
				/* both pass */
				return (board().mycount - board().matchcount);
			}
			return - cp->item(0).think().prefinalize1b(1, INT_MAX);
		}
		/* no room to choose */
		if (board().totalcount<=55) {
			cp->item(0).think().aheadresult
			  = c = - cp->item(0).think().prefinalize1b(0, INT_MAX);
		}else {
			cp->item(0).think().aheadresult
			  = c = - cp->item(0).think().finalize(0, INT_MAX);
		}
		return c;
		/* return - cp->item(0).think().prefinalize1b(0, INT_MAX); */
	}
	
	alleval();
	/* for (l=0; l<cp->number(); l++) cp->item(l).think().eval(); */
	cp->handsort(6);
	
	if (board().totalcount<=55) {
		for (l=0; l<cp->number(); l++)
			cp->item(l).think().aheadresult = INT_MAX;
	}
	
	min = INT_MAX; minpos = -1; c = 0;
	for (l=0; l<cp->number(); l++) {
		if (board().totalcount<=55) {
			cp->item(l).think().aheadresult
			  = c = - cp->item(l).think().prefinalize1b(0, min);
		}else {
			cp->item(l).think().aheadresult
			  = c = - cp->item(l).think().finalize(0, min);
		}
		if (c < min) {
			//if (minpos >= 0) cp->erasehand(minpos);
			min = c;
			minpos = l;
		}else {
			//cp->erasehand(l);
		}
		if (orgmin <= -min) { l++; break; }	// alpha-beta
		if (min<0 /*&& l>=1*/) { l++; break; }
		if (/*board().totalcount <= 55 &&*/ ((orgmin>=0)^!(!board().myflag))) {
			////if (board().totalcount <= 52 && l>=3) break;
			if (board().totalcount >= 53 && l>=3) break;
			if (l>=4) break;
		}else {
			if (l>=5) break;
		}
	}
	
	if (min < 0) {
		for (c=0; c<l; c++) {
			if (cp->item(c).think().aheadresult >= 0) cp->erasehand(c);
		}
	}
	
	if (cp->number() > 7) {
		c = 0;
		for (l=0; l<cp->number(); l++) {
			if (&cp->item(l) == NULL) continue;
			if (c>=7) cp->erasehand(c);
			c++;
		}
	}

	return min;
}

int
taka2Think1::prefinalize2(int passcount, int orgmin, int limit)
{
	int l;
	int c, min, minpos;
	int n/*, cpf*/;

	if (board().totalcount == 64) {
		return (board().mycount - board().matchcount);
	}
	total_breadth++;
	
	/*cpf = !evalflag;*/
	if (cp == NULL) {
		cp = new taka2Choice(board());
		for (l=0; l<cp->number(); l++) {
			cp->item(l).think().lasthands = cp->number();
		}
	}
	if (cp->number() <= 1) {
		if (cp->number() == 0) {
			/* pass */
			if (passcount) {
				/* both pass */
				return (board().mycount - board().matchcount);
			}
			return - cp->item(0).think().prefinalize2(1, INT_MAX, -limit);
		}
		/* no room to choose */
		if (board().totalcount<=55) {
			return - cp->item(0).think().prefinalize2(0, INT_MAX, -limit);
		}else {
			if ((c = cp->item(0).think().aheadresult) == INT_MAX) {
				c = - cp->item(0).think().finalize(0, INT_MAX);
			}
			return c;
		}
		/* return - cp->item(0).think().prefinalize2(0, INT_MAX, -limit); */
	}
		
	alleval();
	/* for (l=0; l<cp->number(); l++) cp->item(l).think().eval(); */
	if (/*cpf ||*/ cp->number() == cp->restnumber()) {
		cp->handsort(6);
	}
	
	min = INT_MAX; minpos = -1; c = 0;
	n = 0;
	for (l=0; l<cp->number(); l++) {
		if (&cp->item(l) == NULL) continue;
		if (board().totalcount<=55) {
			cp->item(l).think().aheadresult
			  = c = - cp->item(l).think().prefinalize2(0, min, -limit);
		}else {
			if (/*cpf ||*/ (c = cp->item(l).think().aheadresult) == INT_MAX) {
				cp->item(l).think().aheadresult
				  = c = - cp->item(l).think().finalize(0, min);
			}
		}
		if (c < min) {
			if (minpos >= 0) cp->erasehand(minpos);
			min = c;
			minpos = l;
		}else {
			cp->erasehand(l);
		}
		if (orgmin <= -min) break;	// alpha-beta
		n++;
		if (((limit>0)^(!board().myflag))) {
			if (n>=4 && init_totalcount<=50 && board().totalcount >= 52) break;
			//if (board().totalcount <= 54 && n>=4) break;
			if (n>=5) break;
		}else {
			if (min<=limit && n>=3) break;
			//if (min<=limit && n>=4) break;
		}
		if (board().totalcount <= 51 && total_breadth > 1300 && n>=2) break;
	}

	return min;
}

int
taka2Think1::semifinalize(int passcount, int orgmin)
{
	int l;
	int c, min, minpos;
	int e;

	if (board().totalcount >= 60) return finalize(passcount, orgmin);
#if TAKADEBUG*0
cout <<"semifinalize() level "
  <<(64 - board().mycount - board().matchcount)
  <<" passcount=" <<passcount <<" myflag=" <<(int)board().myflag
  <<" mycount=" <<(int)board().mycount
  <<" matchcount=" <<(int)board().matchcount <<endl;
#endif
	if (board().totalcount == 64) {
//cout <<"  64filled. return " <<(board().mycount - board().matchcount)
//  <<endl;
		return (board().mycount - board().matchcount);
	}
	total_breadth++;
	
	/*bp = board().create_reverse();*/
	if (cp == NULL) {
		cp = new taka2Choice(board());
		for (l=0; l<cp->number(); l++) {
			cp->item(l).think().lasthands = cp->number();
		}
	}
	if (cp->number() <= 1) {
		if (cp->number() == 0) {
			/* pass */
			if (passcount) {
				// both pass
#if TAKADEBUG*0
cout <<"  bothpass. return " <<(board().mycount - board().matchcount)
  <<endl;
#endif
				return (board().mycount - board().matchcount);
			}
//cout <<"  onepass." <<endl;
			return - cp->item(0).think().semifinalize(1, INT_MAX);
		}
		/* no room to choose */
//cout <<"  no room to choose." <<endl;
		return - cp->item(0).think().semifinalize(0, INT_MAX);
	}
	
#if TAKADEBUG*0
cout <<"  next level. (" <<cp->number() <<")" <<endl;
#endif
	alleval();
	e = cp->number();
	/* for (l=0; l<e; l++) cp->item(l).think().eval(); */
#if 0
	if (e >= 4) {
		e = 4;
		if (board().totalcount > 54) e = 3;
	}
#endif
	if (e >= 3) e = 3;
	cp->handsort(e);
	
	min = INT_MAX; minpos = -1; c = 0;
	for (l=0; l<e; l++) {
#if TAKADEBUG*0
cout <<"recursive level "
  <<(64 - board().mycount - board().matchcount)
  <<" item " <<l <<endl;
#endif
		cp->item(l).think().aheadresult
		  = c
		  = - cp->item(l).think().semifinalize(0, min);
		if (c < min) {
			if (minpos >= 0) cp->erasehand(minpos);
			min = c;
			minpos = l;
		}else {
			cp->erasehand(l);
		}
		if (orgmin <= -min) break;	// alpha-beta
		/*
		if (l==1 && min<0 && board().totalcount<= 54)
			break;
		*/
		if (l==1 && min<0 && init_totalcount<=48) break;
	}
#if TAKADEBUG*0
cout <<"recursive level "
  <<(64 - board().mycount - board().matchcount)
  <<" exit -> " <<minpos <<"<" <<min <<">" <<endl;
//cout <<"     ";
//for (l=0; l<cp->number(); l++) cout <<l <<"<" <<cp->item(l).think().aheadresult <<"> ";
cout <<endl;
#endif

	return min;
}

int
taka2Think1::nthforward(int passcount, int orgmin, int depth)
{
	int l;
	int c, min, minpos;
	int e;

	if (board().totalcount == 64) {
		return (board().mycount - board().matchcount)*10000;
	}
	if (depth == 0) {
		return -weight;
#if 0
		if (board().myflag) return -weight;	/* weight has negative value */
		printf("hoge-\n");
		return (board().mycount - board().matchcount);
#endif
	}
	/* bp = board().create_reverse(); */
	if (cp == NULL) {
		cp = new taka2Choice(board());
		for (l=0; l<cp->number(); l++) {
			cp->item(l).think().lasthands = cp->number();
		}
	}
	if (cp->number() <= 1) {
		alleval();
		if (cp->number() == 0) {
			/* pass */
			if (passcount) {
				// both pass
				return (board().mycount - board().matchcount)*10000;
			}
//cout <<"  onepass." <<endl;
			/* cp->item(0).think().eval(); */
			return - cp->item(0).think().nthforward(1, INT_MAX, depth);
		}
		/* no room to choose */
//cout <<"  no room to choose." <<endl;
		/* cp->item(0).think().eval(); */
		return - cp->item(0).think().nthforward(0, INT_MAX, depth-1);
	}
	
#if TAKADEBUG*0
cout <<"  next level. (" <<cp->number() <<")" <<endl;
#endif
	alleval();
	e = cp->number();
	/* for (l=0; l<e; l++) cp->item(l).think().eval(); */
#if 1
	if (e > 3) {
		if (init_totalcount >= 47) {
			if (board().totalcount > 54) e = 3;
			else e=4;
		}else {
			if (board().totalcount > 43) e = 4;
			else if (e > 5) {
				if (init_totalcount >= 30) e = 5;
				else if (board().totalcount >= 24) e = 6;
				else if (e > 8) e = 8;
			}
		}
	}
#else
	if (e >= 3) e = 3;
#endif
	cp->handsort(e);
	
	min = INT_MAX; minpos = -1; c = 0;
	for (l=0; l<e; l++) {
#if TAKADEBUG*0
cout <<"recursive level "
  <<(64 - board().mycount - board().matchcount)
  <<" item " <<l <<endl;
#endif
		cp->item(l).think().aheadresult
		  = c
		  = - cp->item(l).think().nthforward(0, min, depth-1);
		if (c < min) {
			if (minpos >= 0) cp->erasehand(minpos);
			min = c;
			minpos = l;
		}else {
			cp->erasehand(l);
		}
		if (orgmin <= -min) break;	// alpha-beta
		if (depth < 4 && l==2 && min<0 && (board().totalcount <= 54))
			break;
	}
#if TAKADEBUG*0
cout <<"recursive level "
  <<(64 - board().mycount - board().matchcount)
  <<" exit -> " <<minpos <<"<" <<min <<">" <<endl;
//cout <<"     ";
//for (l=0; l<cp->number(); l++) cout <<l <<"<" <<cp->item(l).think().aheadresult <<"> ";
cout <<endl;
#endif

	return min;
}

/*----------------------------------------------------------------------*/
